import typing
from awsglue.utils import getResolvedOptions
import sys
import dataclasses


def parse_arguments_to_dict(
    custom_argument_keys: typing.Sequence[str],
    builtin_argument_keys: typing.Optional[typing.Sequence[str]] = None,
):
    args = getResolvedOptions(sys.argv, custom_argument_keys)
    all_argument_keys = (
        list(custom_argument_keys) + list(builtin_argument_keys)
        if builtin_argument_keys
        else custom_argument_keys
    )
    return {key: args[key] for key in all_argument_keys}


class Arguments(typing.Protocol):
    """Interface for Glue job argument dataclasses.
    Can be implemented by defining a dataclass that inherrits
    from this protocol.
    The Dataclass can be instantiated by running
    dataclass.from_glue_arguments() insidevan AWS Glue job.
    """

    __dataclass_fields__: typing.ClassVar[typing.Dict]

    @classmethod
    def from_dict(cls, dict: typing.Dict[str, str]):
        return cls(**dict)

    @classmethod
    def _get_builtin_keys(cls):
        builtin_keys = ["JOB_NAME", "JOB_ID", "JOB_RUN_ID"]
        return [
            field_name
            for field_name in cls._field_names()
            if field_name in builtin_keys
        ]

    @classmethod
    def _field_names(cls):
        return [field.name for field in dataclasses.fields(cls)]

    @classmethod
    def from_glue_arguments(cls):
        return cls.from_dict(
            parse_arguments_to_dict(
                cls._field_names(), cls._get_builtin_keys()
            )
        )


@dataclasses.dataclass(kw_only=True)
class BaseJobArguments(Arguments):
    """Implementation of Arguments, with just the builtin
    'JOB_NAME' argument"""

    JOB_NAME: str



from pyspark.context import SparkContext
from pyspark.sql import SparkSession
from awsglue.context import GlueContext
import typing


class SparkEnabledSession(typing.Protocol):
    def get_spark_session(self) -> SparkSession:
        ...


class GlueSession(typing.Protocol):
    def get_glue_context(self) -> GlueContext:
        ...


class SparkGlueSession:
    def __init__(self) -> None:
        self._spark_context = SparkContext()
        self._glue_context = GlueContext(self._spark_context)
        self._spark_session: SparkSession = self._glue_context.spark_session

    def get_spark_session(self):
        return self._spark_session

    def get_glue_context(self):
        return self._glue_context



from pyspark.conf import SparkConf
from pyspark.sql import SparkSession
from awsglue.context import GlueContext  # type: ignore


class HudiGlueSession:
    def __init__(self) -> None:
        self._spark_session: SparkSession = self._set_spark()
        self._spark_context = self._spark_session.sparkContext
        self._glue_context = GlueContext(self._spark_context)

    def get_spark_session(self):
        return self._spark_session

    def _set_spark(self) -> SparkSession:
        conf_list = self._get_conf_list()
        spark_conf = SparkConf().setAll(conf_list)
        return (
            SparkSession.builder.config(conf=spark_conf)
            .enableHiveSupport()
            .getOrCreate()
        )

    def _get_conf_list(self):
        return [
            ("spark.serializer", "org.apache.spark.serializer.KryoSerializer"),
            (
                "spark.sql.catalog.spark_catalog",
                "org.apache.spark.sql.hudi.catalog.HoodieCatalog",
            ),
            (
                "spark.sql.extensions",
                "org.apache.spark.sql.hudi.HoodieSparkSessionExtension",
            ),
        ]



import enum
import typing
from dataclasses import dataclass


class TableType(enum.Enum):
    COPY_ON_WRITE = "COPY_ON_WRITE"
    MERGE_ON_READ = "MERGE_ON_READ"


def get_table_type_opt_key_value(table_type: TableType):
    if table_type == TableType.COPY_ON_WRITE:
        return "COW_TABLE_TYPE_OPT_VAL"
    elif table_type == TableType.MERGE_ON_READ:
        return "MOR_TABLE_TYPE_OPT_VAL"


class IndexType(enum.Enum):
    BLOOM = "BLOOM"
    SIMPLE = "SIMPLE"
    GLOBAL_BLOOM = "GLOBAL_BLOOM"
    GLOBAL_SIMPLE = "GLOBAL_SIMPLE"


class KeyGeneratorNotKnownException(Exception):
    def __init__(self, *args: object) -> None:
        super().__init__(
            (
                "This combination of arguments have not been experimented to see "
                "which key generator should be used. Experiment and update the "
                "relevant tests if in dev and acc this new combination can be make "
                "to work experimentally"
            ),
            *args,
        )


class KeyGenerator(enum.Enum):
    """The keygenerator class has to be specified based on the
    configuration of a table.
    The weird thing is that there is a NonpartitionedKeyGenerator class for
    when you dont wanna partition your data. However, the SimpleKeyGenerator just works
    for non-partitioned data. When mutliple columns are specified for the record key,
    the ComplexKeyGenerator has to be used. But this one apparently cannot handle
    empty (None) partitions.
    https://hudi.apache.org/blog/2021/02/13/hudi-key-generators/

    """

    SIMPLE = "SimpleKeyGenerator"
    COMPLEX = "ComplexKeyGenerator"
    TIMESTAMP = "TimestampBasedKeyGenerator"
    NONPARTITIONED = "NonpartitionedKeyGenerator"
    CUSTOM = "CustomKeyGenerator"


def get_keygenerator_class(
    n_record_key_columns: int,
    partitioning: bool,
    partitioned_on_datetime: bool,
):
    if (
        n_record_key_columns == 1
        and not partitioning
        and not partitioned_on_datetime
    ):
        return KeyGenerator.COMPLEX
    elif (
        n_record_key_columns > 1
        and not partitioning
        and not partitioned_on_datetime
    ):
        return KeyGenerator.COMPLEX
    elif n_record_key_columns > 1 and partitioning and partitioned_on_datetime:
        return KeyGenerator.CUSTOM
    else:
        raise KeyGeneratorNotKnownException(
            "number of record key columns:",
            n_record_key_columns,
            "partitioning:",
            partitioning,
            "partitioned on datetime:",
            partitioned_on_datetime,
        )


def get_keygenerator_class_name(key_generator: KeyGenerator):
    return f"org.apache.hudi.keygen.{key_generator.value}"


@dataclass
class DatetimePartitioning:
    input_date_format: str = "yyyy-MM-dd'T'HH:mm:ss.SSSZ"
    timezone: str = "UTC"
    output_date_format: str = "yyyyMMdd"


def get_additional_options_for_datetime_keygen(
    datetime_partitioning: typing.Optional[DatetimePartitioning],
):
    # https://hudi.apache.org/blog/2021/02/13/hudi-key-generators/#timestampbasedkeygenerator
    if datetime_partitioning:
        return {
            "hoodie.deltastreamer.keygen.timebased.timestamp.type": "DATE_STRING",
            "hoodie.deltastreamer.keygen.timebased.input.dateformat": datetime_partitioning.input_date_format,
            "hoodie.deltastreamer.keygen.timebased.input.dateformat.list.delimiter.regex": "",  # noqa: E501
            "hoodie.deltastreamer.keygen.timebased.input.timezone": "",
            "hoodie.deltastreamer.keygen.timebased.output.dateformat": datetime_partitioning.output_date_format,  # noqa: E501
            "hoodie.deltastreamer.keygen.timebased.timezone": datetime_partitioning.timezone,
        }
    else:
        return {}


def parse_record_key_columns(columns: typing.List[str]):
    if not columns:
        raise ValueError("No columns set for record key")
    return ",".join(columns)


def get_index_type_option(index_type: IndexType):
    return {"hoodie.index.type": index_type.value}


def get_partitioning_options(
    partition_key_column_name: typing.Union[None, str],
    partitioned_on_datetime: bool,
    key_generator_class: KeyGenerator,
):
    if key_generator_class == KeyGenerator.CUSTOM:
        # https://hudi.apache.org/docs/0.12.1/key_generation#customkeygenerator
        partition_key_type = (
            "TIMESTAMP" if partitioned_on_datetime else "SIMPLE"
        )
        partition_field_value = (
            f"{partition_key_column_name}:{partition_key_type}"
        )
    else:
        partition_field_value = (
            partition_key_column_name if partition_key_column_name else ""
        )
    return {
        "hoodie.datasource.write.partitionpath.field": partition_field_value,
        "hoodie.datasource.hive_sync.assume_date_partitioning": "false",
        "hoodie.datasource.write.hive_style_partitioning": "true",
    }


MetaOptions = typing.TypedDict(
    "MetaOptions",
    {"className": str, "hoodie.datasource.write.keygenerator.class": str},
)

CatalogOptions = typing.TypedDict(
    "CatalogOptions",
    {
        "hoodie.table.name": str,
        "hoodie.datasource.hive_sync.table": str,
        "hoodie.datasource.hive_sync.database": str,
        "hoodie.datasource.hive_sync.enable": str,
        "hoodie.datasource.hive_sync.mode": str,
        "hoodie.datasource.hive_sync.use_jdbc": str,
        "hive_sync.support_timestamp": str,
        "path": str,
    },
)

UpsertOptions = typing.TypedDict(
    "UpsertOptions",
    {
        "hoodie.datasource.write.operation": str,
        "hoodie.datasource.write.table_type_opt_key": str,
        "hoodie.datasource.write.recordkey.field": str,
        "hoodie.datasource.write.precombine.field": str,
    },
)

hudi_option_keys = [
    # Meta
    "className",
    "hoodie.datasource.write.keygenerator.class",
    # cataloging
    "hoodie.table.name",
    "hoodie.datasource.hive_sync.table",
    "hoodie.datasource.hive_sync.database",
    "hoodie.datasource.hive_sync.enable",
    "hoodie.datasource.hive_sync.mode",
    "hoodie.datasource.hive_sync.use_jdbc",
    "hive_sync.support_timestamp",
    "path",
    # upsert
    "hoodie.datasource.write.operation",
    "hoodie.datasource.write.table_type_opt_key",
    "hoodie.datasource.write.recordkey.field",
    "hoodie.datasource.write.precombine.field",
    # partitioning
    "hoodie.datasource.write.partitionpath.field",
    "hoodie.datasource.hive_sync.assume_date_partitioning",
    "hoodie.datasource.write.hive_style_partitioning",
    # partitioning - timestamp
    "hoodie.deltastreamer.keygen.timebased.timestamp.type",
    "hoodie.deltastreamer.keygen.timebased.input.dateformat",
    "hoodie.deltastreamer.keygen.timebased.input.dateformat.list.delimiter.regex",  # noqa: E501
    "hoodie.deltastreamer.keygen.timebased.input.timezone",
    "hoodie.deltastreamer.keygen.timebased.output.dateformat",
    "hoodie.deltastreamer.keygen.timebased.timezone",
]

HudiOptionsDict = typing.TypedDict(
    "HudiOptions", {key: str for key in hudi_option_keys}  # type: ignore
)

fixed_hudi_options = {
    "className": "org.apache.hudi",
    "hoodie.datasource.write.operation": "upsert",
    "hoodie.datasource.hive_sync.enable": "true",
    "hoodie.datasource.hive_sync.use_jdbc": "false",
    "hoodie.datasource.hive_sync.mode": "hms",
    "hive_sync.support_timestamp": "true",
}


def get_hudi_options(
    database_name: str,
    table_name: str,
    table_type: TableType,
    datetime_partitioning: typing.Optional[DatetimePartitioning],
    index_type: IndexType,
    record_key_columns: typing.List[str],
    precombine_column_name: str,
    partition_key_column_name: typing.Optional[str],
    hudi_table_path: str,
):
    if not partition_key_column_name:
        partition_key_column_name = ""
    key_generator_class = get_keygenerator_class(
        len(record_key_columns),
        True if partition_key_column_name else False,
        True if datetime_partitioning else False,
    )
    options: typing.Dict[str, str] = {
        **fixed_hudi_options,
        "hoodie.table.name": table_name,
        "hoodie.datasource.write.table_type_opt_key": get_table_type_opt_key_value(
            table_type
        ),
        "hoodie.datasource.write.keygenerator.class": get_keygenerator_class_name(
            key_generator_class
        ),
        "hoodie.datasource.write.recordkey.field": parse_record_key_columns(
            record_key_columns
        ),
        "hoodie.datasource.write.precombine.field": precombine_column_name,
        "hoodie.datasource.hive_sync.enable": "true",
        "hoodie.datasource.hive_sync.database": database_name,
        "hoodie.datasource.hive_sync.table": table_name,
        "path": hudi_table_path,
        **get_index_type_option(index_type),
        **get_partitioning_options(
            partition_key_column_name,
            True if datetime_partitioning else False,
            key_generator_class,
        ),
        **get_additional_options_for_datetime_keygen(datetime_partitioning),
    }

    return options



import dataclasses
from glue_helper_lib import table
from glue_helper_lib.hudi import config
import typing
import pyspark.sql
import enum


@dataclasses.dataclass
class Partitioning:
    partition_column: str
    datetime: typing.Optional[config.DatetimePartitioning]


class WriteMode(enum.Enum):
    OVERWRITE = "overwrite"
    UPSERT = "upsert"


@dataclasses.dataclass
class WriteHudiTableArguments:
    catalog: table.GlueCatalogArguments
    index_type: config.IndexType
    table_type: config.TableType
    record_key_colums: typing.List[str]
    precombine_column: str
    partitioning: typing.Optional[Partitioning]
    write_mode: WriteMode


class HudiGlueTable:
    _write_mode_to_mode_string = {
        WriteMode.OVERWRITE: "overwrite",
        WriteMode.UPSERT: "append",
    }

    def __init__(self, storage_location: table.StorageLocation) -> None:
        self._storage_location = storage_location

    def _get_hudi_config(self, write_args: WriteHudiTableArguments):
        if not write_args.partitioning:
            datetime_partitioning = None
            partition_key_column_name = None
        else:
            datetime_partitioning = write_args.partitioning.datetime
            partition_key_column_name = (
                write_args.partitioning.partition_column
            )

        return config.get_hudi_options(
            database_name=write_args.catalog.database,
            table_name=write_args.catalog.table,
            hudi_table_path=str(self._storage_location),
            table_type=write_args.table_type,
            datetime_partitioning=datetime_partitioning,
            index_type=write_args.index_type,
            record_key_columns=write_args.record_key_colums,
            precombine_column_name=write_args.precombine_column,
            partition_key_column_name=partition_key_column_name,
        )

    def write(
        self, df: pyspark.sql.DataFrame, write_args: WriteHudiTableArguments
    ):
        df.write.format("hudi").options(
            **self._get_hudi_config(write_args)
        ).mode(self._write_mode_to_mode_string[write_args.write_mode]).save()

    def read(
        self, spark_session: pyspark.sql.SparkSession
    ) -> pyspark.sql.DataFrame:
        return spark_session.read.format("hudi").load(
            str(self._storage_location)
        )



import logging
import typing
import enum
import sys


class LogLevel(enum.Enum):
    DEBUG = logging.DEBUG
    INFO = logging.INFO
    WARNING = logging.WARNING
    ERROR = logging.ERROR
    CRITICAL = logging.CRITICAL


class Logger(logging.Logger):
    """Logger for AWS glue that puts the logs into the 'output'
    cloudwatch logs of a Glue job.

    Args:
        name: name of the logger
        level: minimum log level of log statements that are logged
        formatter: Format for the log messages: defaults to
            <time - name - [level] - message>
    """

    def __init__(
        self,
        name: str,
        level: LogLevel,
        formatter: typing.Optional[logging.Formatter] = None,
    ) -> None:
        super().__init__(name, level.value)
        self._configure(formatter)

    def _configure(self, formatter: typing.Optional[logging.Formatter]):
        handler = logging.StreamHandler(sys.stdout)
        if not formatter:
            formatter = logging.Formatter(
                "%(asctime)s - %(name)s - [%(levelname)s] - %(message)s"
            )
        handler.setFormatter(formatter)
        self.addHandler(handler)



import dataclasses
import typing
import pyspark.sql


class S3Uri:
    def __init__(self, path: str) -> None:
        self._path = path
        if not path.startswith("s3://"):
            raise ValueError("path should start with 's3://'")

    def __str__(self) -> str:
        return self._path


class GlueTable(typing.Protocol):
    def write(self, df: pyspark.sql.DataFrame):
        ...

    def read(
        self, spark_session: pyspark.sql.SparkSession
    ) -> pyspark.sql.DataFrame:
        ...


@dataclasses.dataclass
class GlueCatalogArguments:
    database: str
    table: str


class StorageLocation:
    def __init__(self, path: S3Uri):
        self._path = path

    def __str__(self) -> str:
        return str(self._path)


class ParquetTable:
    def __init__(
        self,
        storage_location: StorageLocation,
        partition_columns: typing.Optional[typing.List[str]] = None,
        compression: typing.Optional[str] = None,
    ) -> None:
        super().__init__()
        self._storage_location = storage_location
        self._partition_columns = partition_columns
        self._compression = compression

    def write(self, df: pyspark.sql.DataFrame):
        df.write.parquet(
            str(self._storage_location),
            mode="overwrite",
            partitionBy=self._partition_columns,
            compression=self._compression,
        )

    def read(
        self, spark_session: pyspark.sql.SparkSession
    ) -> pyspark.sql.DataFrame:
        return spark_session.read.parquet(str(self._storage_location))
